from abc import ABC, abstractmethod
from typing import List, Any


class BaseDBHandler(ABC):
    @abstractmethod
    def init_db(self):
        pass

    @abstractmethod
    def get_item(self, item_id: int, item_type: type('Item')) -> 'item_type':
        pass

    @abstractmethod
    def get_items_query(self, item_type: type('Item')) -> Any:
        pass

    @abstractmethod
    def add_item(self, item_data: 'Item'):
        pass

    @abstractmethod
    def add_items(self, items: List['Item']):
        pass

    @abstractmethod
    def update_item(self, item_data: 'Item'):
        pass

    @abstractmethod
    def delete_item(self, item_id: int, item_type: type('Item')):
        pass

    @abstractmethod
    def delete_items(self, item_ids: List[int], item_type: type('Item')):
        pass

    @abstractmethod
    def rollback(self):
        pass

    @abstractmethod
    def commit(self):
        pass

    @abstractmethod
    def reset(self):
        pass



from flask_api import status
from flask_restful import abort

from .db_handlers import db
from .helper_models.item import Item


def get_item_or_abort(item_id: str, item_type: type(Item)) -> Item:
    item: Item = db.get_item(item_id, item_type)
    if not item:
        abort(status.HTTP_404_NOT_FOUND, error=f'Item of type {item_type.__name__} with id {item_id} not found.')

    return item



from ...db_utils.helper_models.item import Item
from ...db_utils.queries import get_item_or_abort
from marshmallow import fields


class ItemByIdField(fields.Int):
    def __init__(self, item_type: type('Item'), **kwargs):
        super(ItemByIdField, self).__init__(**kwargs)
        self.__item_type = item_type

    def _deserialize(self, value, attr, data, **kwargs) -> Item:
        return get_item_or_abort(value, self.__item_type)




from typing import List

from marshmallow import Schema, EXCLUDE, ValidationError
from flask import request
from werkzeug.exceptions import BadRequest


def unpack_not_none_to_dict(value: any, target: dict):
    if value is not None:
        return {**target, **value}

    return target


class BaseSchema(Schema):
    """
    Meta.strict True means a validation error is thrown if something doesn't fit.
    Meta.unknown Exclude means that instead of throwing an error on bonus data, the bonus data is just ignored.
    """
    class Meta:
        strict = True
        unknown = EXCLUDE


def get_data_groups() -> List:
    data_groups: List = []
    try:
        data_groups.append(request.json)
    except BadRequest:
        pass

    try:
        data_groups.append(request.values)
    except BadRequest:
        pass

    try:
        data_groups.append(request.files)
    except BadRequest:
        pass

    return data_groups


def load_schema(schema, unpack=True):
    """
    Taking the parameters from request json, view_args, args by this order and passing it to the schema for validation.
    :param schema: The resource schema to validate
    :param unpack: Should or shouldn't return values as validate_data or ever value in its own key in the kwargs
    """
    def wrap(f):
        def decorator(*args, **kwargs):
            try:
                data = {}
                data_groups = [kwargs, *get_data_groups()]

                for data_item in data_groups:
                    data = unpack_not_none_to_dict(data_item, data)

                schema_instance: dict = schema().load(data=data)

                if unpack:
                    kwargs.update(**schema_instance)
                else:
                    kwargs.update({'valid_data': schema_instance})
                return f(*args, **kwargs)
            except ValidationError as err:
                return {'error': err.messages}, 400

        return decorator

    return wrap



from abc import ABC, abstractmethod
from typing import Union

from ..db_handlers import global_session


class AbstractSqlAlchemyMetaClass(type(ABC), type(global_session.Model)):
    pass


class Item(ABC, global_session.Model, metaclass=AbstractSqlAlchemyMetaClass):
    __abstract__ = True

    id: Union[int, global_session.Column] = global_session.Column('Id', global_session.Integer, primary_key=True,
                                                                  autoincrement=True)

    def __init__(self, item_id: int):
        self.id = item_id

    @abstractmethod
    def _serialize(self, *args, **kwargs) -> dict:
        pass

    def serialize(self, *args, **kwargs) -> dict:
        return {
            'id': self.id,
            **self._serialize(*args, **kwargs)
        }



from .schema_validation import BaseSchema


class SerializationSchema(BaseSchema):
    __data: dict

    def __init__(self, **kwargs):
        super().__init__()
        self.__data = {}
        for field in self.fields:
            self.__data[field] = kwargs.pop(field)

    def serialize(self) -> dict:
        return self.dump(self.__data)


from abc import ABC, abstractmethod
from typing import Type

from ..db_utils.db_handlers import db
from ..db_utils.helper_models.item import Item


class SerializerError(Exception):
    pass


class BaseSerializer(ABC):
    _model: Item

    def __init__(self, model: Item = None, model_id: int = None):
        self._model = model

        if not model:
            if model_id:
                self._model = db.get_item(model_id, self._model_type)
            else:
                raise SerializerError('Serializer received neither a model nor an id.')

    @property
    @abstractmethod
    def _model_type(self) -> Type[Item]:
        pass

    @abstractmethod
    def _fetch_sub_models(self):
        pass

    @abstractmethod
    def _make_dict(self) -> dict:
        pass

    def serialize(self) -> dict:
        self._fetch_sub_models()
        return self._make_dict()



from typing import List
from flask import current_app
from flask_sqlalchemy import SQLAlchemy
from sqlalchemy.orm import Query

from .base_db_handler import BaseDBHandler


class SqlDBHandler(BaseDBHandler):
    db: SQLAlchemy

    def __init__(self):
        from . import global_session
        self.db = global_session

    @property
    def session(self):
        return self.db.session

    def init_db(self):
        self.db.init_app(current_app)
        self.db.create_all()

    def get_item(self, item_id: int, item_type: type('Item')) -> 'item_type':
        return self.session.query(item_type).filter_by(id=item_id).scalar()

    def get_items_query(self, item_type: type('Item')) -> Query:
        return self.session.query(item_type)

    def add_item(self, item_data: 'Item'):
        self.session.add(item_data)

    def add_items(self, items: List['Item']):
        self.session.add_all(items)

    def update_item(self, item_data: 'Item'):
        raise NotImplementedError('Have not implemented update item')

    def delete_item(self, item_id: int, item_type: type('Item')):
        self.session.query(item_type).filter_by(id=item_id).delete()

    def delete_items(self, item_ids: List[int], item_type: type('Item')):
        for item_id in item_ids:
            self.delete_item(item_id, item_type)

    def rollback(self):
        self.session.rollback()

    def commit(self):
        self.session.commit()

    def reset(self):
        pass



from setuptools import setup, find_packages

setup(
    name='atlas_flask_utils',
    version='1.1.0',
    author='Venfi Oranai',
    author_email='venfioranai@gmail.com',
    package_dir={'': 'src'},
    packages=find_packages('src'),
    install_requires=[
        'flask',
        'marshmallow',
        'werkzeug',
        'sqlalchemy',
        'flask_sqlalchemy',
        'flask_api',
        'flask_restful'
    ]
)



from ...db_utils.queries import get_item_or_abort
from marshmallow.validate import Validator


class ExistingItemValidator(Validator):
    def __init__(self, item_type: type('Item')):
        self.item_type = item_type

    def __call__(self, value: str):
        get_item_or_abort(value, self.item_type)
